;Notes:
;Animation
;This template stores the animation keys of a specific frame. 
;It should contain at least one AnimationKey template. It should also contain a reference to the target frame. 
AnimationSet
Acts as a container for Animation templates. The Animation templates contained in this set have the same time values. 


;=======================================================================================
;These PROTOS needed to be defined before the functions can be used, since the functions
;are used in THIS file, and since THIS file is included before the procedures have been defined.
;=======================================================================================
FrameNode_FindFrame                        PROTO :DWORD,:DWORD
FrameNode_SetTransformationMatrix PROTO :DWORD,:DWORD
;=======================================================================================
;These structs are the in-memory versions as used in our database
;=======================================================================================
	BONEPOSITIONKEY STRUCT	
		dwTime dd ?
		Pos D3DXVECTOR3 <>
	BONEPOSITIONKEY ENDS

	BONESCALEKEY STRUCT
		dwTime dd ?
		vScale D3DXVECTOR3 <>
	BONESCALEKEY ENDS

	BONEROTATIONKEY STRUCT
		dwTime dd ?
		Rotation D3DXQUATERNION <>
	BONEROTATIONKEY ENDS

	BONEMATRIXKEY STRUCT	
		dwTime dd ?
		Mat D3DXMATRIX <>
	BONEMATRIXKEY ENDS

LPBONEPOSITIONKEY typedef ptr BONEPOSITIONKEY 
LPBONESCALEKEY typedef ptr BONESCALEKEY 
LPBONEROTATIONKEY typedef ptr BONEROTATIONKEY 
LPBONEMATRIXKEY typedef ptr BONEMATRIXKEY 
;=======================================================================================
;These structs are the originals as found in the XFILE
;=======================================================================================
	tmpPOSITIONKEY STRUCT
		dwTime DWORD ?
		dwFloatsCount DWORD ?
		Pos D3DXVECTOR3 <>
	tmpPOSITIONKEY ENDS

	tmpSCALEKEY STRUCT
		dwTime DWORD ?
		dwFloatsCount DWORD ?
		Scale D3DXVECTOR3 <>
	tmpSCALEKEY ENDS

	tmpROTATIONKEY STRUCT
		dwTime DWORD ?
		dwFloatsCount DWORD ?
;		//NOTE x files are w x y z and QUATERNIONS are x y z w
		w FLOAT ?
		x FLOAT ?
		y FLOAT ?
		z FLOAT ?
	tmpROTATIONKEY ENDS


	tmpMATRIXKEY STRUCT
		dwTime DWORD ?
		dwFloatsCount DWORD ?
		Mat D3DXMATRIX <>
	tmpMATRIXKEY ENDS


LPFRAMENODE typedef ptr FrameNode
;====================================================================================
AnimationNode STRUCT
            LO LinkedObject <>
	pPositionKeys   LPBONEPOSITIONKEY ?    		;< The array of position keys
	pScaleKeys LPBONESCALEKEY ?			;< The array of scale keys
	pRotationKeys LPBONEROTATIONKEY ?		;< The array of rotation keys
	pMatrixKeys LPBONEMATRIXKEY ?			;< The array of matrix keys
	dwPositionKeyCount dd ?        			;< The size of the position keys array
	dwScaleKeyCount dd ?				;< The size of the scale keys array
	dwRotationKeyCount dd ?        			;< The size of the rotation keys array
	dwMatrixKeyCount dd ?				;< The size of the matrix keys array
;           -----------------------------------
	pOwnersName	 LPBYTE ?			            ;< The name of the target frame
	pTargetFrame LPFRAMENODE ?			;< A pointer to the target frame
AnimationNode ENDS
LPANIMATIONNODE typedef ptr AnimationNode
;==========================================================================================
;==========================================================================================
;==========================================================================================
;==========================================================================================
; Load data from the x file. 
;The data loaded are the animation keys of this animation object.
;  pxofobj = The x file data object to load from
;==========================================================================================
- Check the type of the data object
  -If the type is a reference
    Get the referenced template, which is a frame template
    Get the name of it
    Store the name
   
  -Else if type is data
		Check the type of the animation key
		Load the key accordingly
  -Endif

;==========================================================================================
AnimationNode_Load PROC pThis:LPANIMATIONNODE, pxofobj:DWORD
;==========================================================================================
	local pxofdataobj:DWORD
	local pxofrefobj:DWORD
	local objecttype:LPGUID
	local pdata:ptr
            local protationkey, ppositionkey,ptranslationkey,pscalekey,pmatrixkey:ptr
 
	local dwsize, dwkeytype, dwkeyscount, dwkey, dwnamelength:DWORD
	local szframename[256]:BYTE
;==========================================================================================
            mov esi,pThis
            assume esi:ptr AnimationNode

;	// Query the child for it's FileDataReference
	mcall [pxofobj], IDirectXFileObject_QueryInterface, addr IID_IDirectXFileDataReference, addr pxofrefobj 
	.if eax== SUCCEEDED
;		// The object is of type reference. Use it to get a reference to the targetframe
		mcall [pxofrefobj],IDirectXFileDataReference_Resolve, addr pxofdataobj 
		.if eax== SUCCEEDED		
;			// The original data is found
			mcall [pxofdataobj],IDirectXFileData_GetType,addr objecttype 
			.if eax== FAILED			
				_saferelease pxofrefobj
				_saferelease pxofdataobj
				Errr CTXT  ("Couldn't retrieve object type while loading animation")
			.endif

;			// The object -MUST- be a frame
                                    .if $invoke (CmpGUID, objecttype, ADDR  TID_D3DRMFrame,16) == TRUE
;				// The frame is found, get its name
;				// The name will be used later by the FindBone function to get
;				// a pointer to the target frame

				.if [esi].pTargetFrame != NULL
					_saferelease pxofrefobj
					_saferelease pxofdataobj
					Errr CTXT ( "Animation frame name reference duplicated" )
				.endif

;				// get the length of the name
				mcall [pxofdataobj],IDirectXFileData_GetName,NULL, addr dwnamelength
				.if eax== FAILED
					_saferelease pxofrefobj
					_saferelease pxofdataobj
					Errr CTXT( "Couldn't retrieve frame's name while loading animation")
				.endif

				.if dwnamelength == 0 
					_saferelease pxofrefobj
					_saferelease pxofdataobj
					Errr CTXT( "No name found for an animation frame" )
				.endif

;				// get the name
				mcall [pxofdataobj],IDirectXFileData_GetName,addr szframename, addr dwnamelength
				.if eax== FAILED				
					_saferelease pxofrefobj
					_saferelease pxofdataobj
					Errr CTXT( "Couldn't retrieve frame's name while loading animation")
				.endif
	;			pOwnersName = szframename;
			.endif
			_saferelease pxofrefobj
		.endif 
		_saferelease pxofrefobj

	.else
;		// Query the child for it's FileData
		mcall [pxofobj],IDirectXFileObject_QueryInterface, addr IID_IDirectXFileData, addr pxofdataobj 
		.if eax== SUCCEEDED		
;			// a data object is found, get its type
			mcall [pxofdataobj],IDirectXFileData_GetType,addr objecttype
			.if eax== FAILED
				_saferelease pxofdataobj
				Errr CTXT( "Couldn't retrieve object type")
			.endif

                  .if $invoke (CmpGUID, objecttype, ADDR  TID_D3DRMAnimationKey ,16) == TRUE
;				// an animation key is found , so load the data
				mcall [pxofdataobj],IDirectXFileData_GetData, 0, addr dwsize, addr pdata 
				.if eax== FAILED				
					_saferelease pxofdataobj
					Errr CTXT ( "Couldn't retrieve data")
				.endif

;				// get the type and count of the key
                                                mov edi,pdata
                                                m2m dwkeytype ,     dword ptr [esi]             ;=dword [0]
				m2m dwkeyscount , dword ptr[esi+4]          ;=dword [1]
;---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
				.if dwkeytype == 0                          ;/*rotation key*/ 				
					.if [esi].pRotationKeys != NULL 					
						_saferelease pxofdataobj
						Errr CTXT( "Rotation key duplicated" )
					.endif
                                                            mov ebx,sizeof BONEROTATIONKEY
                                                            mov eax,dwkeyscount
                                                            mul ebx                                                        
					mov [esi].pRotationKeys,$invoke (GlobalAlloc,GPTR,EAX)   ;allocate BoneRotationKeys Array
					.if [esi].pRotationKeys==NULL					
						_saferelease pxofdataobj
						Errr CTXT ("Out of memory allocating BoneRotationKeys")
					.endif
					m2m [esi].dwRotationKeyCount , dwkeyscount

;					//NOTE x files are w x y z and QUATERNIONS are x y z w
                                                            mov ebx,edi
                                                            add ebx,8
					mov protationkey , ebx

;					for( dwkey = 0; dwkey < dwkeyscount; dwkey++ )
                                                            xor eax,eax
                                                            mov dwkey,eax
                                                            .while eax<dwkeyscount
                                                                 mov ebx,sizeof BONEROTATIONKEY
                                                                 mul ebx
                                                                  mov edi,[esi].pRotationKeys         
                                                                 add edi,eax                                 ;edi=pRotationKeys[dwkey]
                                                                  mov ebx,protationkey     
                                                                  assume  ebx:nothing 
                                                                  assume ebx:ptr BONEROTATIONKEY
                                                                  assume edi:ptr BONEROTATIONKEY
        				                  m2m [edi].dwTime , [ebx].dwTime
        				                  m2m [edi].Rotation.x , [ebx].Rotation.x
        				                  m2m [edi].Rotation.y , [ebx].Rotation.y
        				                  m2m [edi].Rotation.z , [ebx].Rotation.z
        				                  m2m [edi].Rotation.w , [ebx].Rotation.w                    
        				                  add protationkey,sizeof BONEROTATIONKEY        
                                                                  inc dwkey
                                                                  mov eax,dwkey
				            .endw
;---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
				.elseif dwkeytype == 1                                          ;/*scale key*/ 				
					.if [esi].pScaleKeys != NULL					
						_saferelease pxofdataobj
						Errr CTXT( "Scale key duplicated" )
					.endif
                                                            mov ebx,sizeof BONESCALEKEY
                                                            mov eax,dwkeyscount
                                                            mul ebx  
                                                      
					mov [esi].pScaleKeys ,$invoke (GlobalAlloc,GPTR,EAX)   ;allocate BoneScaleKeys Array
					.if [esi].pScaleKeys==NULL 					
						_saferelease pxofdataobj
						Errr CTXT ("Out of memory allocating BoneScaleKeys")
					.endif
					m2m [esi].dwScaleKeyCount , dwkeyscount

                                                             mov ebx,pdata
                                                            add ebx,8
					mov pscalekey , ebx

                                                          xor eax,eax
                                                            mov dwkey,eax
					.while eax<dwkeyscount
                                                                        mov ebx,sizeof BONESCALEKEY
                                                                        mul ebx
                                                                        mov edi,[esi].pScaleKeys
                                                                        add edi,eax
                                                                        mov ebx,pscalekey
                                                                      assume edi:ptr BONESCALEKEY
                                                                      assume ebx:ptr tmpSCALEKEY
			     			m2m [edi].dwTime , [ebx].dwTime
						invoke RtlMoveMemory, addr [edi].vScale ,addr [ebx].Scale, sizeof BONESCALEKEY.vScale
                                                           add pscalekey,sizeof BONESCALEKEY
                                                          inc dwkey
                                                          mov eax,dwkey
					.endw
;---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
				.elseif dwkeytype == 2                                              ;/*position key*/ 
					.if [esi].pPositionKeys != NULL					
						_saferelease pxofdataobj
						Errr CTXT ( "Position key duplicated" )
					.endif

                                          mov ebx,sizeof BONEPOSITIONKEY
                                          mov eax,dwkeyscount
                                          mul ebx    
					mov [esi].pPositionKeys ,$invoke (GlobalAlloc,GPTR,EAX)   ;allocate BonePositionKeys Array
					.if !eax 
						_saferelease pxofdataobj
						Errr CTXT ("Out of memory allocating BonePositionKeys")
					.endif
					m2m [esi].dwPositionKeyCount , dwkeyscount

                                           mov ebx,pdata
                                           add ebx,8
					mov ppositionkey , ebx
                                                xor eax,eax
					mov dwkey,eax
                                                .while eax<dwkeyscount
                                                          mov ebx,sizeof BONEPOSITIONKEY
                                                          mul ebx
                                                          mov edi,[esi].pPositionKeys
                                                          add edi,eax
                                                          mov ebx,ppositionkey 
                                                          assume ebx:ptr tmpPOSITIONKEY
                                                          assume edi:ptr BONEPOSITIONKEY
					    m2m [edi].dwTime , [ebx].dwTime
					    m2m [edi].Pos.x , [ebx].Pos.x
				              m2m [edi].Pos.y , [ebx].Pos.y
				              m2m [edi].Pos.z , [ebx].Pos.z
					    add ppositionkey, sizeof BONEPOSITIONKEY
                                                          inc dwkey
                                                          mov eax,dwkey
				      .endw
;---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
				.elseif dwkeytype == 4                                      ;/*matrix key*/ 				
					.if [esi].pMatrixKeys != NULL 					
						_saferelease pxofdataobj
						Errr CTXT( "Matrix key duplicated" )
					.endif
                
                                         mov ebx,sizeof BONEMATRIXKEY
                                         mov eax,dwkeyscount
                                         mul ebx    
			     mov [esi].pMatrixKeys ,$invoke (GlobalAlloc,GPTR,EAX)   ;allocate BoneMatrixKeys Array
			     .if !eax 
				_saferelease pxofdataobj
				Errr CTXT ("Out of memory allocating BoneMatrixKeys")
			      .endif
			      m2m [esi].dwMatrixKeyCount , dwkeyscount

                                      mov ebx,pdata
                                      add ebx,8
			      mov pmatrixkey , ebx   ; pmatrixkey = (tmpMATRIXKEY*)(pdata + (sizeof(DWORD) * 2));
			      xor eax,eax
                                          mov dwkey,eax
			      .while eax<dwkeyscount
                                                   mov ebx,sizeof BONEMATRIXKEY
                                                   mul ebx
                                                   mov edi,[esi].pMatrixKeys
                                                   add edi,eax
                                                   mov ebx,pmatrixkey
				m2m [edi].dwTime , [ebx].dwTime
                                                assume edi:nothing
                                                assume edi:LPBONEMATRIXKEY
                                                assume ebx:nothing
                                                assume ebx:ptr tmpMATRIXKEY
				invoke RtlMoveMemory, addr [edi].Mat , addr [ebx].Mat, sizeof D3DXMATRIX                    
				add pmatrixkey, sizeof BONEMATRIXKEY

                                                   inc dwkey
                                                   mov eax,dwkey
					.endw
;---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
				.else					;*** the type is unknown, report the error
						_saferelease pxofdataobj
					       Errr CTXT ( "Unexpected (and unhandled) animation key type" )
				.endif

			.endif
			_saferelease pxofdataobj 
		.endif
	.endif
     return S_OK
AnimationNode_Load ENDP
;==========================================================================================
;==========================================================================================
;==========================================================================================
;==========================================================================================
AnimationNode_Destroy PROC pThis:LPANIMATIONNODE
            .if pThis==NULL
                Errr CTXT ("MORON Tried to destroy NULL AnimationNode")
            .endif
            push esi
            mov esi,pThis
            assume esi:LPANIMATIONNODE
	_safefree  [esi].pMatrixKeys
	_safefree [esi].pPositionKeys 
	_safefree  [esi].pRotationKeys 
	_safefree  [esi].pScaleKeys 
	.while [esi].LO.pChild !=NULL                
                invoke AnimationNode_Destroy, [esi].LO.pChild   ;nested call to destroy children
           .endw
           invoke KillEntry, esi
           pop esi
            ret
AnimationNode_Destroy ENDP
;==========================================================================================
;==========================================================================================
;==========================================================================================
; Find the target bone of this animation.
;pThis= The Animation Node we are referring to (the subject or target node of this procedure)
;prootframe= The root frame from which to start the search for the bone
;==========================================================================================
AnimationNode_FindBone PROC  pThis:LPANIMATIONNODE, prootframe:LPFRAMENODE
            .if prootframe==NULL
                Errr CTXT ("MORON called AnimationNode_FindBone with NULL rootframe pointer")
            .endif
            .if pThis==NULL
                Errr CTXT ("MORON called AnimationNode_FindBone with NULL AnimationNode")
            .endif
            push esi
            mov esi,pThis
            assume esi:LPANIMATIONNODE
	.if [esi].pOwnersName != NULL
		m2m [esi].pTargetFrame , $invoke (FrameNode_FindFrame, prootframe, [esi].pOwnersName)
            .endif
;	// find the bones of the child objects
	.while [esi].LO.pChild!=0
		invoke AnimationNode_FindBone, esi, prootframe    ;Nested call to deal with Children
		mov esi,[esi].LO.pChild
	.endw
            pop esi
            ret
AnimationNode_FindBone ENDP
;==========================================================================================
;==========================================================================================
;==========================================================================================
;Animate the bones according to the given time.
 ;dwtime = The requested time.
;==========================================================================================
AnimationNode_SetTime PROC pThis:LPANIMATIONNODE, dwtime:DWORD
;==========================================================================================
local ikey, ikey1, ikey2,dwtime1,dwtime2, temp1, temp2:DWORD
local resultmatrix, tempmatrix :D3DXMATRIX 
local flerpvalue:FLOAT
local scalevector:D3DXVECTOR3 
local positionvector:D3DXVECTOR3 
local quat:D3DXQUATERNION 
local banimate:BOOL
local q1:D3DXQUATERNION 
local q2:D3DXQUATERNION 

.if pThis==NULL
    Errr CTXT("MORON Tried to set time for a NULL AnimationNode")
.endif
mov banimate,FALSE
push esi
mov esi,pThis
assume esi:LPANIMATIONNODE
;	// if no target frame is found, then this object doesn't hold animation keys
	.if [esi].pTargetFrame!=NULL	
;		// check the type of the key
		.if [esi].pMatrixKeys!=NULL	
                                    mov ebx,[esi].pMatrixKeys
                                    mov eax,[esi].dwMatrixKeyCount
                                    dec eax
                                    mov ecx,sizeof BONEMATRIXKEY
                                    mul ecx
                                    add ebx,eax
                                    assume ebx:ptr BONEMATRIXKEY
			    m2m dwtime,  [ebx].dwTime

;			// get the two keys between which the time is currently in
;			for( ikey = 0 ; ikey < dwMatrixKeyCount; ikey++ )
                                    xor eax,eax
                                    mov ikey,eax
                                    .while eax<[esi].dwMatrixKeyCount
                                                mov ebx,[esi].pMatrixKeys
                                                mov ecx,sizeof BONEMATRIXKEY
                                                mul ecx
                                                add ebx,eax

                                           mov eax,dwtime
				.if [ebx].dwTime > eax 				
					m2m ikey2 , ikey
					.if ikey > 0 
                                                                        mov eax,ikey
                                                                        dec eax
						mov ikey1,eax
					.else                                           ;// when ikey == 0, then ikey1 == 0
						m2m ikey1 , ikey
					            .break
                                                            .endif
				.endif
                                                inc ikey
                                                mov eax,ikey
                                    .endw    

                                    mov eax,ikey1
                                    mov ebx,[esi].pMatrixKeys
                                    mov ecx,sizeof BONEMATRIXKEY
                                    mul ecx
                                    add ebx,eax
		            m2m dwtime1 , [ebx].dwTime
                                    mov eax,ikey2
                                    mov ebx,[esi].pMatrixKeys
                                    mov ecx,sizeof BONEMATRIXKEY
                                    mul ecx
                                    add ebx,eax
			m2m dwtime2 , [ebx].dwTime


;			// get the lerp value from the time
                                    mov eax,dwtime2
                                    sub eax,dwtime1
			.if eax ==0
                                                fld fp0 
				fstp flerpvalue
			.else
                                                mov eax,dwtime
                                                sub eax,dwtime1
                                                mov ebx,dwtime2
                                                sub ebx,dwtime1
                                                mov eax,temp1
                                                mov eax,temp2
                                                fild temp1
                                                fidiv temp2
                                                fstp flerpvalue		;flerpvalue = float(dwtime-dwtime1) / float(dwtime2-dwtime1);
                                    .endif
;			// we cannot apply a lerp function on two matrices, so just get the nearest one
                                fld fphalf
                                fld flerpvalue
                                fcom
                                __FJLE @F
			m2m ikey , ikey2
                                jmp nofix
@@:			m2m ikey , ikey1
nofix:
;			// apply the transformation
                                    mov eax,ikey
                                    imul eax,sizeof BONEMATRIXKEY
                                    add eax,[esi].pMatrixKeys
                                    mov EBX ,[esi].pTargetFrame
			invoke FrameNode_SetTransformationMatrix, EBX , addr [eax].BONEMATRIXKEY.Mat


;-------------------------------------------------------------------------------------------------------------------------------------------------------------
		.else
;			// initialize a matrix to hold the result of the transformation
			invoke D3DXMatrixIdentity,addr resultmatrix

;			// check the scale key
			.if [esi].pScaleKeys!=NULL
				mov ikey1,0
                                                mov ikey2,0			
                                                mov eax,[esi].dwScaleKeyCount
                                                dec eax
                                                imul eax, sizeof BONESCALEKEY
                                                add eax,[esi].pScaleKeys
				m2m dwtime , [eax].BONESCALEKEY.dwTime

;				// get the two keys between which the time is currently in
;				for( ikey = 0; ikey < dwScaleKeyCount; ikey++ )
                                                xor eax,eax
                                                mov ikey,eax
                                                .while eax<[esi].dwScaleKeyCount
                                                            imul eax, sizeof BONESCALEKEY
                                                            add eax,[esi].pScaleKeys
                                                      mov ebx,dwtime
					.if [eax].BONESCALEKEY.dwTime > ebx 
						m2m ikey2 , ikey
						.if ikey > 0 
                                                                                    mov eax,ikey
                                                                                    dec eax
							mov ikey1,eax
						.else                                     ;  // when ikey == 0, then dwp2 == 0
							m2m ikey1 , ikey
						              .break
                                                                        .endif
                                                            .endif
                                                            inc ikey
                                                            mov eax,ikey
                                                .endw
                                                mov eax,ikey1
                                                imul eax, sizeof BONESCALEKEY
                                                add eax,[esi].pScaleKeys
				m2m dwtime1 , [eax].BONESCALEKEY.dwTime
                                                mov eax,ikey2
                                                imul eax, sizeof BONESCALEKEY
                                                add eax,[esi].pScaleKeys
				m2m dwtime2 ,  [eax].BONESCALEKEY.dwTime

;				// get the lerp value
                                                mov eax,dwtime2
                                                sub eax,dwtime1
				.if eax == 0 
                                                      fld fp0
					fstp flerpvalue
				.else
                                                        mov eax,dwtime
                                                        sub eax,dwtime1
                                                         mov temp1,eax
                                                         mov eax,dwtime2
                                                         sub eax,dwtime1
                                                         mov temp2,eax
                                                         fild temp1
                                                         fidiv temp2
                                                         fstp flerpvalue                    ;=  float(dwtime-dwtime1)/float(dwtime2-dwtime1);
                                                .endif
;				// apply the lerp function on the scale vector
;				D3DXVec3Lerp( &scalevector, &pScaleKeys[ikey1].Scale,&pScaleKeys[ikey2].Scale, flerpvalue );
                                                mov eax,ikey1
                                                mov ebx,ikey2
                                                imul eax,sizeof BONESCALEKEY
                                                imul ebx,sizeof BONESCALEKEY
                                                add eax,[esi].pScaleKeys
                                                add ebx,[esi].pScaleKeys         
                                                assume ebx:nothing
                                                lea ecx,scalevector
                                                __LoadFloat3 [eax].BONESCALEKEY.vScale
                                                __LerpFloat3 [ebx].BONESCALEKEY.vScale, flerpvalue
                                                __StoreFloat3 scalevector


;				// create the matrix
				invoke D3DXMatrixScaling, addr tempmatrix, scalevector.x, scalevector.y, scalevector.z 
				invoke D3DXMatrixMultiply, addr resultmatrix, addr resultmatrix, addr tempmatrix 
				mov banimate ,TRUE
			.endif
                            ;-----------------------------------------------------------------------------------------------------------------------------------------
;			//check rot keys
			.if [esi].pRotationKeys != NULL			
				mov ikey1,0
                                                mov ikey2,0
                                                mov eax,[esi].dwRotationKeyCount
                                                dec eax
                                                imul eax,sizeof BONEROTATIONKEY
                                                add eax,[esi].pRotationKeys
				m2m dwtime ,[eax].BONEROTATIONKEY.dwTime

;				// get the two keys surrounding the current time value    				
                                                xor eax,eax
                                                mov ikey,eax
                                                .while eax<[esi].dwRotationKeyCount             ;for( ikey = 0; ikey < dwRotationKeyCount; ikey++ )
                                                            imul eax,sizeof BONEROTATIONKEY
                                                            add eax,[esi].pRotationKeys
                                                      mov ebx,dwtime
					.if [eax].BONEROTATIONKEY.dwTime > ebx 
						m2m ikey2 , ikey
						.if ikey > 0 
                                                                                    mov eax,ikey
                                                                                    dec eax
							mov ikey1 , eax
						.else                                             ;// when ikey == 0, then dwp2 == 0
							m2m ikey1 , ikey
						             .break
                                                                 .endif
					.endif
                                                      inc ikey
                                                      mov eax,ikey
                                                .endw
                                                mov eax,ikey1
                                                imul eax,sizeof BONEROTATIONKEY
                                                add eax,[esi].pRotationKeys
				      m2m dwtime1 , [eax].BONEROTATIONKEY.dwTime
                                                mov eax,ikey2
                                                imul eax,sizeof BONEROTATIONKEY
                                                add eax,[esi].pRotationKeys
				m2m dwtime2 ,  [eax].BONEROTATIONKEY.dwTime

;				// get the lerp value
                                                mov eax,dwtime2
                                                sub eax,dwtime1
				.if eax == 0 
                                                            fld fp0
					fstp flerpvalue 
				.else
                                                        mov eax,dwtime
                                                        sub eax,dwtime1
                                                        mov temp1,eax
                                                        mov eax,dwtime2
                                                        sub eax,dwtime1
                                                        mov temp2,eax
                                                        fild temp1
                                                        fidiv temp2
                                                        fstp flerpvalue                 ; = float(dwtime-dwtime1)/float(dwtime2-dwtime1);
                                                .endif

;				//apply spherical lerp function
;				q1.x = -pRotationKeys[ikey1].Rotation.x;
;				q1.y = -pRotationKeys[ikey1].Rotation.y;
;				q1.z = -pRotationKeys[ikey1].Rotation.z;
;				q1.w = pRotationKeys[ikey1].Rotation.w;

;				q2.x = -pRotationKeys[ikey2].Rotation.x;
;				q2.y = -pRotationKeys[ikey2].Rotation.y;
;				q2.z = -pRotationKeys[ikey2].Rotation.z;
;				q2.w = pRotationKeys[ikey2].Rotation.w;

                                                mov eax,ikey1
                                                imul eax,sizeof BONEROTATIONKEY
                                                add eax,[esi].pRotationKeys
                                                lea edi,q1
                                                assume edi:ptr D3DXQUATERNION
                                                fld [eax].BONEROTATIONKEY.Rotation.w
                                                fld [eax].BONEROTATIONKEY.Rotation.z
                                                fld [eax].BONEROTATIONKEY.Rotation.y
                                                fld [eax].BONEROTATIONKEY.Rotation.x
                                                fchs
                                                fstp [edi].x
                                                fchs
                                                fstp [edi].y
                                                fchs
                                                fstp [edi].z
                                                fstp [edi].w 

                                                mov eax,ikey2
                                                imul eax,sizeof BONEROTATIONKEY
                                                add eax,[esi].pRotationKeys
                                                lea edi,q2
                                                assume edi:ptr D3DXQUATERNION
                                                fld [eax].BONEROTATIONKEY.Rotation.w
                                                fld [eax].BONEROTATIONKEY.Rotation.z
                                                fld [eax].BONEROTATIONKEY.Rotation.y
                                                fld [eax].BONEROTATIONKEY.Rotation.x
                                                fchs
                                                fstp [edi].x
                                                fchs
                                                fstp [edi].y
                                                fchs
                                                fstp [edi].z
                                                fstp [edi].w 

				invoke D3DXQuaternionSlerp, addr quat, addr q1, addr q2, flerpvalue 

;				// create the matrix
				invoke D3DXMatrixRotationQuaternion, addr tempmatrix, addr quat 
				invoke D3DXMatrixMultiply, addr resultmatrix, addr resultmatrix, addr tempmatrix 
				mov banimate , TRUE
			.endif
                            ;-----------------------------------------------------------------------------------------------------------------------------------------
;			// check for position keys
			.if [esi].pPositionKeys!=NULL			
				mov ikey1,0
                                                mov ikey2,0
                                                mov eax,[esi].dwPositionKeyCount
                                                dec eax
                                                imul eax,sizeof BONEPOSITIONKEY
                                                add eax,[esi].pPositionKeys
				m2m dwtime , [eax].BONEPOSITIONKEY.dwTime

;				// get the two keys surrounding the time value
                                                xor eax,eax
                                                mov ikey,eax
                                                .while eax<[esi].dwPositionKeyCount
                                                            imul eax,sizeof BONEPOSITIONKEY
                                                            add eax,[esi].pPositionKeys
                                                      mov ebx,dwtime
					.if [eax].BONEPOSITIONKEY.dwTime > ebx 					
						m2m ikey2 , ikey1
						.if ikey > 0
                                                                                    mov eax,ikey
                                                                                    dec eax 
							mov ikey1,eax 
						.else                                         ; // when ikey == 0, then dwp2 == 0
							m2m ikey1 , ikey
						             .break
                                                                         .endif
					.endif
                                                            inc ikey
                                                            mov eax,ikey
                                                .endw
                                                mov eax,ikey1
                                                imul eax,sizeof BONEPOSITIONKEY
                                                add eax,[esi].pPositionKeys
				m2m dwtime1 , [eax].BONEPOSITIONKEY.dwTime
                                                mov eax,ikey2
                                                imul eax,sizeof BONEPOSITIONKEY
                                                add eax,[esi].pPositionKeys
				m2m dwtime2 ,[eax].BONEPOSITIONKEY.dwTime

;				// get the lerp value
                                                mov eax,dwtime2
                                                sub eax,dwtime1
				.if eax == 0 
                                                            fld fp0
					fstp flerpvalue
				.else
                                                        mov eax,dwtime
                                                        sub eax,dwtime1
                                                        mov temp1,eax
                                                        mov eax,dwtime2
                                                        sub eax,dwtime1
                                                        mov temp2,eax
                                                        fild temp1
                                                        fidiv temp2
                                                        fstp flerpvalue                 ; = float(dwtime-dwtime1)/float(dwtime2-dwtime1);
                                                .endif

;				// apply the lerp function
;				D3DXVec3Lerp( (D3DXVECTOR3*)&positionvector, &pPositionKeys[ikey1].Pos, &pPositionKeys[ikey2].Pos, flerpvalue );
                                                mov eax,ikey1
                                                mov ebx,ikey2
                                                imul eax,sizeof BONEPOSITIONKEY
                                                imul ebx,sizeof BONEPOSITIONKEY
                                                add eax,[esi].pPositionKeys
                                                add ebx,[esi].pPositionKeys
                                                __LoadFloat3 [eax].BONEPOSITIONKEY.Pos
                                                __LerpFloat3  [ebx].BONEPOSITIONKEY.Pos, flerpvalue 
                                                __StoreFloat3 positionvector


;				// prepare the matrix
				invoke D3DXMatrixTranslation, addr tempmatrix, positionvector.x, positionvector.y, positionvector.z 
				invoke D3DXMatrixMultiply, addr resultmatrix, addr resultmatrix, addr tempmatrix 
				mov banimate , TRUE

			.else
;				// if no position key is found, use the original position of the frame
                                                mov esi, [esi].pTargetFrame
                                                assume esi:ptr FrameNode
                                                lea eax,[esi].OriginalMatrix
				invoke D3DXMatrixTranslation, addr tempmatrix, \
                                                                                [eax].D3DXMATRIX._41, \
                                                                                [eax].D3DXMATRIX._42, \
                                                                                [eax].D3DXMATRIX._43 
				invoke D3DXMatrixMultiply, addr resultmatrix, addr resultmatrix, addr tempmatrix 
			.endif
;-----------------------------------------------------------------------------------------------------------------------------------------
			.if banimate==TRUE
                                                assume esi:LPANIMATIONNODE
                                                invoke FrameNode_SetTransformationMatrix, [esi].pTargetFrame, addr resultmatrix
                                    .endif
		.endif
	.endif

;	// call the settime of the children
	.while [esi].LO.pChild !=NULL	
		invoke AnimationNode_SetTime, [esi].LO.pChild, dwtime 		
	.endw
            return S_OK
AnimationNode_SetTime ENDP
;==========================================================================================
;==========================================================================================
;==========================================================================================
;==========================================================================================
