
SkinMesh_LoadAnimationSet PROTO :DWORD,:DWORD
SkinMesh_LoadAnimation      PROTO :DWORD,:DWORD

;=============================================================================
FrameNode STRUCT
	LO LinkedObject <>
	pMeshes LPMESHNODE ?		            ;	< Meshes in this Frame
	OriginalMatrix D3DXMATRIX <NULL>		;	< The matrix originally loaded with the frame
	TransformationMatrix D3DXMATRIX <NULL>	;	< The transformation of the frame
; This matrix is the one that's updated at each rendering operation.
; It's obtained by combining the matrix of this frame with the matrix 
; of the parent frame. */
	CombinedMatrix D3DXMATRIX <>       	;            < The combined matrix
FrameNode ENDS
LPFRAMENODE typedef ptr FrameNode

SkinMesh STRUCT
    pFramesRoot       LPFRAMENODE ?             ;Root of entire frame hierarchy of this SkinMesh Model
    pAnimationsRoot LPANIMATIONNODE ?       ;Root of entire animation set of this SkinMesh Model
SkinMesh ENDS
LPSKINMESH typedef ptr SkinMesh


;=============================================================================
;This procedure parses all the "File Objects" in a given File Object, loading them into a Frame Hierarchy.
;If it finds instances of a "Frame Object" while doing so, it will iteratively call itself.
;Noting that it performs an Append of the new FrameNode onto the existing one,
;what happens if we find more than one "Frame Object" within a Frame?
;------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
include FrameNode_Load.inc

;=============================================================================
;=============================================================================
;=============================================================================
FrameNode_Destroy PROC pThis:LPFRAMENODE
.if pThis==NULL
    Errr CTXT ("MORON Tried to Destroy NULL FrameNode")
.endif
mov esi,pThis
assume esi:LPFRAMENODE
.if [esi].LO.pChild !=NULL
    invoke FrameNode_Destroy,[esi].LO.pChild
.endif
invoke MeshNode_Destroy, [esi].pMeshes
pop esi
ret
FrameNode_Destroy ENDP
;=============================================================================
;=============================================================================
;=============================================================================
; Compute the world matrix of the frame by combining 
;the transformation matrix of this frame with the parent's matrix
; pmat : A pointer to the parent's world matrix

FrameNode_Update PROC pThis:LPFRAMENODE, pmat:LPD3DXMATRIX 
    .if pThis==NULL 
        ret
    .endif
    push esi
    mov esi,pThis
    assume esi:LPFRAMENODE
    invoke D3DXMatrixMultiply, addr [esi].CombinedMatrix, addr [esi].TransformationMatrix, pmat 
;// update child frames
    .if [esi].LO.pChild !=0
          mov esi,[esi].LO.pChild
          .while esi!=0
  	    invoke FrameNode_Update, [esi].LO.pChild, addr [esi].CombinedMatrix
	    mov esi,[esi].LO.pChild
          .endw
    .endif
    pop esi
ret
FrameNode_Update ENDP
;=============================================================================
;=============================================================================
;=============================================================================


FrameNode_Render PROC pThis:LPFRAMENODE
            assume esi:nothing
            .if pThis==NULL                                                 ;Theres only one input param, and its REQUIRED !!
                ret
            .endif
            push esi
            mov esi,pThis                                   ;If we have child meshes, set the world matrix and rend them
            assume esi:LPFRAMENODE
	.if [esi].pMeshes != NULL
                push esi
      	    mcall [lpd3dDevice],IDirect3DDevice8_SetTransform, 256, addr [esi].CombinedMatrix 
                assume esi:LPFRAMENODE
    	    mov esi,[esi].pMeshes       	;Fetch RootPointer of this Frame's MeshNodes linkedlist
                assume esi:LPMESHNODE
    	    .while esi!=NULL	
    		invoke MeshNode_Render, esi       ;Render any and all the MeshNodes for this FrameNode
    		mov esi,[esi].LO.pChild
    	    .endw
                pop esi 
            .endif

           .if [esi].LO.pChild!=NULL                    ;If we have child frames, iteratively rend them
                  mov esi,[esi].LO.pChild
      	     .while esi!=0	
      	         invoke FrameNode_Render,esi
      	         mov esi,[esi].LO.pChild
                 .endw
	.endif
            pop esi
            return S_OK
FrameNode_Render ENDP
;=============================================================================
;=============================================================================
;=============================================================================
; Find a frame in this frame hierarchy with the specified name.

FrameNode_FindFrame PROC pThis:DWORD, pName:DWORD
        .if pThis==NULL || pName==NULL 
            Errr CTXT ("FrameNode_FindFrame requires BOTH params, neither can be NULL !!")
        .endif
        push esi
        mov esi,pThis
        assume esi:LPFRAMENODE
        mov eax,[esi].LO.pName
        .if pName == eax                                                    ;Quickie test - Are the stringpointers an exact match?
	return pThis
        .else
            invoke lstrcmp,pName,[esi].LO.pName             ;Generic test - Are the two actual strings an exact match?
            .if eax==0
                return pThis
            .endif
        .endif
;		// loop through the child objects
	.while [esi].LO.pChild != NULL		
	     invoke FrameNode_FindFrame ,[esi].LO.pChild, pName
	      .break .if eax!=NULL
	      mov esi,[esi].LO.pChild
	.endw
            pop esi
	ret
FrameNode_FindFrame ENDP
;=============================================================================
;=============================================================================
;=============================================================================
; Find the bones of the child meshes of a frame.
; prootframe = The root of the frame hierarchy to search in for the bones.

FrameNode_FindBones PROC pThis:LPFRAMENODE, prootframe:LPFRAMENODE
local ErrBuf[256]:BYTE
.if prootframe==NULL
    Errr CTXT ("MORON Tried to Find Bones of a NULL FrameNode")
.endif
push esi
mov esi,pThis
assume esi:nothing
invoke wsprintf,addr ErrBuf,CTXT("Trying to find Bones within Frame:%s"),[esi].LinkedObject.pName
invoke MessageBox,0,addr ErrBuf,CTXT("FrameNode_FindBones"),MB_OK
;	// find the bones of the meshes
            mov ebx,pThis
            push ebx
            assume ebx:LPFRAMENODE
            mov esi, [ebx].pMeshes
            assume esi:LPMESHNODE            
            .while esi!=NULL
		invoke MeshNode_FindBones, esi ,prootframe
		mov esi,[esi].LO.pChild
	.endw
            pop esi
;	// find the bones in the child frames
            assume esi:LPFRAMENODE
	mov esi,[esi].LO.pChild
	.while esi!=NULL	
		invoke FrameNode_FindBones,esi, prootframe
		mov esi,[esi].LO.pChild
	.endw
pop esi
ret
FrameNode_FindBones ENDP
;=============================================================================
;=============================================================================
;=============================================================================
;=============================================================================
;This is just a helper to set the TransformationMatrix for a given Frame
;=============================================================================
FrameNode_SetTransformationMatrix PROC pFrame:DWORD, pmat:DWORD
.if pFrame==NULL 
    Errr CTXT("MORON Tried to Set Transformation Matrix on a NULL Frame")
.endif
push esi
mov esi,pFrame
assume esi:LPFRAMENODE
m2m [esi].TransformationMatrix, pmat
pop esi
ret
FrameNode_SetTransformationMatrix ENDP


;=============================================================================
FrameNode_InitObject PROC
invoke D3DXMatrixIdentity, addr [esi].TransformationMatrix 
invoke D3DXMatrixIdentity, addr [esi].OriginalMatrix 
ret
FrameNode_InitObject ENDP
